/**
 * $Id: Actor.java,v 1.40 2003/04/09 20:53:38 deselt Exp $
 *
 * This is the main Actor class.
 */

package salsa.language;

import java.io.IOException; 
import salsa.language.exceptions.*; 
import salsa.messaging.*; 
import salsa.naming.*; 
import salsa.naming.exceptions.*;

public class Actor implements java.io.Serializable {
	protected UAN uan;
	protected UAL ual;

	private transient TheaterService theaterService;
	private transient TransportService transportService;
	private transient NamingService namingService;

	/*
	 * These new constructors have replaced getReferenceByName and getReferenceByLocation
	 * They prevent messy actorState creation and syntactically make more sense.
	 * By default, the compiler uses the Actor(UAL) constructor, as all actors begin
	 * as anonymous actors, and become non-anonymous when they are bound to a UAN
	 */

	public Actor() {
		theaterService = ServiceFactory.getTheater();
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();

		uan = null;
		ual = null;
	}

	public Actor(UAN _uan) {
		theaterService = ServiceFactory.getTheater();
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();

		uan = _uan;
		ual = null;
	}

	public Actor(UAL _ual) {
		theaterService = ServiceFactory.getTheater();
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();

		uan = null;
		ual = _ual;
	}


	/*
	 * These methods retrieve the UAN and UAL of the actor
	 */
	public UAN getUAN() {
		return uan;
	}

	public UAL getUAL() {
		if ( (uan != null) && (theaterService.getEntry(uan) instanceof UAL)) {
			return (UAL)theaterService.getEntry(uan);
		} else {
			return ual;
		}
	}


	/* 
	 * The following cover local and remote message sending for Actors
	 */
	public void send(Message message) {
		//If this actor has been sent a bind message, its UAN needs to be updated
		//so it can be serialized and access its state on the remote theater,  move the
		//states registry entry to the non-anonymous registry, and update the
		//anonymous registry with the uan

		message.messageSent = System.currentTimeMillis();
		
		//if the actor is local, get its state and put the message
		//in its mailbox.
		if (getActorState() != null) {
			synchronized (getActorState().mailbox) {
				//local message send
				if (getActorState() != null) {
					getActorState().send(message);	
					return;
				}
			}
		}

		//remote message send:
		//if the actor is non-anonymous(it has a UAN), get its UAL from the registry
		//if not, remote send the message to the specified UAL

		if (uan == null) transportService.send(message, ual);
		else {
			UAL remoteLocation = (UAL)theaterService.getEntry(uan);
			transportService.send(message, remoteLocation);
		}
	}



	/*
	 * This retrives the actorstate from whichever registry it resides in
	 * if the actor is not anonymous (it has a UAN), it will get the
	 * state from the non-anonymous registry.  If the actor is foreign,
	 * this registry returns the apropriate ual
	 * If the actor is anonymous, it will get the state from the anonymous
	 * registry. If that actor has since become non-anonymous (through the bind method)
	 * that registry will return the UAN, and this reference will become non-anonymous
	 */
	public ActorState getActorState() {
		Object state = null;

		if (uan != null) {
			state = theaterService.getEntry(uan);
		} else if (ual != null) {
			state = theaterService.getEntry(ual);
			if (state instanceof UAN) {
				uan = (UAN) state;
				state = theaterService.getEntry(uan);
			}
		}

		if (state instanceof ActorState) return (ActorState)state;
		else return null;
	}



	/*
	 * Methods for printing out the identity of an actor
	 */
	public String toString() {
		return getClass().getName()+":"+super.toString();
	}

	public String getName() {
		return getActorState().getName();
	}



	/*
	 * Methods for Actor serializability
	 */
	private void writeObject(java.io.ObjectOutputStream out) throws SalsaException {
		//sending reference to local anonymous actor
		//need to keep the theater alive now since the reference has been
		//sent to the rest of the theater system.
		getActorState(); //update uan and ual links

		if (uan == null) {
			RunTime.receivedUniversalActor();
		}

		try {
			out.writeObject(uan);
			out.writeObject(ual);
			out.flush();
		} catch (IOException e) {
			throw new SalsaWrapperException( "Exception during actor reference serialization.", e);
		}
	}

	private void readObject(java.io.ObjectInputStream in) throws SalsaException {
		try {
			uan = (UAN)in.readObject();
			ual = (UAL)in.readObject();
		} catch (Exception e) {
			throw new SalsaWrapperException( "Exception during actor reference deserialization.", e);
		}

		//get new instances of the services used by the Actor Reference
		theaterService = ServiceFactory.getTheater();
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();
	}
}
